package farm.ants.agent;
import java.util.*;

/**
 * Represents a single entry in an agents objective function.
 * <p>
 * Created: Mon Apr  9 20:18:28 2001 
 * <p>
 * Migrated from ANTS to farm Jul 14 2002
 *
 * @author Roger Mailler
 */
public class ObjEntry implements Comparable, Cloneable{
    int agents = 0;
    int slots = 0;

    /**
     * Constructor
     */
    public ObjEntry(){
    }
    
    /**
     * Constructor
     * @param a The number of agents desired
     * @param s The number of slots wanted
     */    
    public ObjEntry(int a, int s) {
	agents = a;
	slots = s;
    }

    /**
     * Constructor
     * @param obj A string version of the ObjEntry
     */
    public ObjEntry(String obj){
	this.fromString(obj);
    }

    /**
     * Gets the number of agents desired by this entry
     * @return The number of agents
     */
    public int numAgents(){return agents;}

    /**
     * Gets the number of slots desired by this entry
     * @return The number of slots
     */
    public int numSlots(){return slots;}

    /**
     * Converts the entry to its string version
     * @return The string version
     */
    public String toString(){return agents + "*" + slots;}

    /**
     * Gets the number of agents desired by this entry
     * @param a  The number of agents
     */
    public void setAgents(int a){agents = a;}

    /**
     * Sets the number of slots desired by this entry
     * @param s The number of slots
     */
    public void setSlots(int s){slots = s;}

    /**
     * Parses and sets the entry from a string version
     * @param obj The string to parse
     */
    public void fromString(String obj){
	StringTokenizer tmpToken =
	    new StringTokenizer(obj, "*");
	//get the new level
	agents = Integer.parseInt(tmpToken.nextToken());
	slots = Integer.parseInt(tmpToken.nextToken());
    }

    /**
     * Tests to see if two ObjEntries are the same
     * @param obj The other ObjEntry
     * @return True if they are equal
     */
    public boolean equals(Object obj){
	ObjEntry objE = (ObjEntry) obj;
	return (objE.numAgents() == agents && objE.numSlots() == slots);
    }

    /**
     * Compares two ObjEntries
     * @param obj The other ObjEntry
     * @return -1 if this is better than o1, 0 if the same, 1 otherwise
     */
    public int compareTo(Object obj){
	double val = ((Math.sqrt(((ObjEntry)obj).numAgents())*((ObjEntry)obj).numSlots())- (Math.sqrt(this.agents)*this.slots));
	if(val < 0.0)
	    return -1;
	else if (val == 0.0)
	    return 0;
	else
	    return 1;

    }

    /**
     * Creates a copy of the entry
     * @return The clone
     */
    public Object clone(){
	return new ObjEntry(agents, slots);
    }
} // ObjEntry
