package farm.ants.agent; 
import java.util.*;
import utilities.Log;
import farm.metagent.*;

/**
 * The recieving end of a SPAM negotiation.  This machine is spawned
 * when an Evaluate Solutions (ES) message is recieved by a SPAM agent
 * <p>
 * Created: Thu Aug 30 15:29:30 2001
 * <p>
 * Migrated from ANTS to farm 07/16/2002
 *
 * @author Roger Mailler
 */
public class SPAMR extends PulseActionMachine {

    PACLock lock;
    String name;
    Agent me = null;

    /**
     * Constructor
     * @param m The agent this SPAM belongs to
     */
    public SPAMR(Agent a) {
	me = a;
	setWaitMess(1);
	setWaitFor(WF_MESSAGES);
	setAction("SPAMR");
 	name = me.getAgentName();
   }

    public boolean executeState(int stateNum){
	switch (stateNum){
	case 0:
	    return executeState0();
	case 1:
	    return executeState1();
	case 2:
	    return executeState2();
	default:
	    log.log("Unknown SPAMR state code:" + stateNum, 
	    	    log.LOG_ERR);
	    return true;
	}
    } 

    boolean executeState0(){
	log.log("Executing SPAMR State 0 "+ getID(), log.LOG_INFO);
	Vector messages = (Vector) getAttribute("Message Queue");
	Track tmpTrack = (Track) getAttribute("Track");
	PulseActionMachine parent  = (PulseActionMachine) getAttribute("Parent");

	PACLock l = new PACLock(getID());
	ObjEntry obj = new ObjEntry();
	Message m = (Message) messages.elementAt(0);
	Vector solutions = parseES(m,obj,l);
	int timeout =  ((Integer) getAttribute("Timeout")).intValue();
	String otherMan = m.getSource();
       	int time = me.getTime();

	if(time > timeout){
	    //crap, timed out
	    log.log("SPAMR Timed Out because the orginating message had a timeout time that has expired",log.LOG_INFO);
	    return true;
	}

	setAttribute("Solutions", solutions);
	setAttribute("Objective", obj);
	setAttribute("Other", otherMan);

	//need the lock
	Vector locks = new Vector();
	locks.add(l);
	lock = l;
	transition(1,WF_NONE, 0, 0, locks);	
	log.log("Requesting lock from Controller " + lock, log.LOG_INFO);

	return false;
    }
    boolean executeState1(){
	log.log("Executing SPAMR State 1 "+ getID(), log.LOG_INFO);
	Track tmpTrack = (Track) getAttribute("Track");
	Vector sensors = tmpTrack.getContext(); 

	Vector solutions = (Vector) getAttribute("Solutions");
	ObjEntry newObjEntry = (ObjEntry) getAttribute("Objective");
	String otherMan = (String) getAttribute("Other");
	int timeout = ((Integer) getAttribute("Timeout")).intValue(); 
	String replyAddr = (String) getAttribute("replyAddr"); 

	if(!Controller.hasLock(otherMan, getID())){
	    log.log("SPAMR Timed Out while waiting for the lock", log.LOG_INFO);
	    Controller.unlock(lock);
	    return true;
	}
	

	//create the list of non-conflicted sensors
	Vector nonConSens = new Vector();
	Hashtable oneSol = (Hashtable) solutions.elementAt(0);
	for(int i=0;i<sensors.size();i++){
	    if(!oneSol.containsKey(sensors.elementAt(i))){
		nonConSens.add(((SensorData) sensors.elementAt(i)).getName());
	    }
	}
	
	//ask the non-conflicting sensors for their schedules
	Hashtable sensScheds = tmpTrack.getSchedules();
	for(int i=0;i<nonConSens.size();i++){
	    if(!sensScheds.containsKey(nonConSens.elementAt(i))){
		sensScheds.put(nonConSens.elementAt(i), new SlotSchedule());
	    }
	}
	for(Iterator i = sensScheds.keySet().iterator();i.hasNext();){
	    String name = (String) i.next();
	    if(nonConSens.indexOf(name)== -1){
		i.remove();
	    }
	}

	setAttribute("nonConScheds",sensScheds);

	//then using the two pieces of information, sort the
	//solutions and return the answer
	Vector orders = orderAlternatives(solutions, sensScheds, newObjEntry);


	String msg = "ES ";
	for(int i=0;i<orders.size()-1;i++){
	    msg += ((Integer) orders.elementAt(i)).toString() + ",";
	}
	msg += ((Integer) orders.elementAt(orders.size() -1)).toString();

	msg = replyAddr + ":" + msg+ ":P" + getID();

	Message m = new Message(name, otherMan, msg);
	me.sendMessage(m);

	log.log("Sending message " + m, log.LOG_INFO);
	transition(2, WF_MESSAGES|WF_TIME, 1, timeout);
	return false;
    }

    boolean executeState2(){
	log.log("Executing SPAMR State 2 "+ getID(), log.LOG_INFO);
	Track tmpTrack = (Track) getAttribute("Track");
 	Vector sensors = tmpTrack.getContext(); 
	Vector solutions = (Vector) getAttribute("Solutions");
	ObjEntry newObjEntry = (ObjEntry) getAttribute("Objective");
	String otherMan = (String) getAttribute("Other");
	Vector messages = (Vector) getAttribute("Message Queue");

	if(messages.size() <1){
	    log.log("SPAMR Timed Out while waiting for the AC message", log.LOG_INFO);
	}
	else{
	    Message m = (Message) messages.elementAt(0);
	    //get the info and remove the PAM address
	    String data = m.getData().substring(m.getData().indexOf(":")+1);
	    //now remove the message type ID
	    data = data.substring(3);

	    //have the number
	    int solNum = Integer.parseInt(data);
	    
	    Hashtable chosenSolution = (Hashtable) solutions.elementAt(solNum);
	    Hashtable currSolution = generateLocalBinding(chosenSolution, tmpTrack.getSchedules(), sensors, newObjEntry);

	    log.log("My final allocation for obj level " + newObjEntry + " will be "+ currSolution.toString(), log.LOG_INFO); 
	    
	    //clear the schedule cache for the newly mediated over sensors
	    for(Iterator i = chosenSolution.keySet().iterator();i.hasNext();){
		tmpTrack.clearSchedule((String) i.next());
	    }

	    bindSolution(tmpTrack, currSolution, newObjEntry);
	}
	Controller.unlock(lock);
	return true;
    }



    Hashtable generateLocalBinding(Hashtable partSolns, Hashtable schedules, Vector sensors, ObjEntry obj){
    
	Hashtable finalSolution = new Hashtable();

	//begin by finding out how many sensors I am given for each slot 
	int givenSlots[] = new int[SlotSchedule.NUMSLOTS];
	for(Iterator i = partSolns.keySet().iterator();i.hasNext();){
	    String sensName = (String) i.next();
	    SlotSchedule SS = (SlotSchedule) partSolns.get(sensName);
	    for(int j=0;j<SlotSchedule.NUMSLOTS;j++){
		if(!SS.isEmpty(j)){
		    //I'm suppose to use this slot
		    givenSlots[j] += 1;
		}
	    }
	    finalSolution.put(sensName, SS);
	}

	//so we have the counts of what is given, let's figure out what
	//is needed from the non-cons
	int needSlots[] = new int[SlotSchedule.NUMSLOTS];
	for(int i=0;i<SlotSchedule.NUMSLOTS;i++){
	    int need = obj.numAgents()-givenSlots[i];
	    needSlots[i] = ((need > sensors.size()-partSolns.size())?0:need);
	}	

	//make sure all of the non-cons show up in the final answer
	Vector nonCon = new Vector();
	for (Iterator i = sensors.iterator();i.hasNext();){
	    String sens = ((SensorData) i.next()).getName();
	    if(!finalSolution.containsKey(sens)){
		finalSolution.put(sens, new SlotSchedule());
		nonCon.add(sens);
	    }
	}



	try{

	for(int i=0;i<SlotSchedule.NUMSLOTS;i++){
	    int have = givenSlots[i];	    
	    int need = needSlots[i];

	    String taking[] = new String[need];
	    int cons[] = new int[need];
	    int worstCon = -1;
	    int worstConPos = -1;
 
	    for(int j = 0;j<need;j++){
		String name = (String) nonCon.elementAt(j);
		SlotSchedule SS = (SlotSchedule) schedules.get(name);
		int conflicts = 0;
		if(SS!=null){
		    conflicts = SS.numCommitments(i);
		    if(SS.hasCommitment(i,name)){
			conflicts--;
		    }
		}
		taking[j] = name;
		cons[j] = conflicts;
		if(conflicts>worstCon){
		    worstCon = conflicts;
		    worstConPos =  j;
		}
	    }
	    for(int j=need;j<nonCon.size();j++){
		String name = (String) nonCon.elementAt(j);
		SlotSchedule SS = (SlotSchedule) schedules.get(name);
		int conflicts = 0;
		if(SS!=null){
		    conflicts = SS.numCommitments(i);
		    if(SS.hasCommitment(i,name)){
			conflicts--;
		    }
		}

		if(conflicts<worstCon){
		    taking[worstConPos] = name;
		    cons[worstConPos] = worstConPos;
		    //find the new bad one
		    worstCon = -1;
		    worstConPos = -1;
		    for(int k = 0;k<need;k++){
			int confs = cons[k];
			if(confs>worstCon){
			    worstCon = confs;
			    worstConPos = k;
			}
		    }
		}
	    }
	    for(int j = 0;j<need;j++){
		SlotSchedule SS = (SlotSchedule) finalSolution.get(taking[j]);
		SS.add(i,name);

	    }
	
	}
	} catch(ArrayIndexOutOfBoundsException AO){
	    String needData = "[";
	    for(int i = 0;i<needSlots.length;i++){
		needData = needData + needSlots[i];
		if(i<needSlots.length -1) needData = needData + ", ";
	    }
	    needData = needData + "]";

	    log.log("Objective: " + obj.toString() + "\n Avail sensors: " + sensors.toString() + "\n NeedSlots " +  needData + "\n nonCons " + nonCon.toString(),Log.LOG_ERR);
	    throw AO;	    
	}
	return finalSolution;
    }
	
    /**
     * The ES request parser
     */
    Vector parseES(Message m, ObjEntry newObjEntry, PACLock l){
	/*this function takes a properly formatted ES message and parses
	  it into a vector of the solutions being explored

	  ES 4*3:S1,S2:1166:P123:lk 407,5537:tm 415

	*/
	String content = m.getData();
	String sender = m.getSource();
	StringTokenizer tok = new StringTokenizer(content, ":");
	
	String newObj = tok.nextToken().substring(3);//get's the objective function
	newObjEntry.fromString(newObj);
	String header = tok.nextToken().trim();//get's the header 
	                                       //for the solutions
	
	//need the soultion header so we can reconstruct the solutions
	Vector names = new Vector();
	StringTokenizer headST = new StringTokenizer(header, ",");
	while(headST.hasMoreElements()){
	    names.add(headST.nextToken());
	}


	//now get the actual solutions
	Vector solutions = new Vector();
	String solns = tok.nextToken().trim();
	int pointer = 0;
	while(pointer < solns.length()){
	    String thisSoln = solns.substring(pointer, pointer + names.size());
	    Hashtable soln = new Hashtable();
	    for(int i=0;i<names.size();i++){
		String sensName = (String) names.elementAt(i);
		int val = Integer.parseInt(thisSoln.substring(i,i+1));
		SlotSchedule SS = new SlotSchedule();
		SS.fromInt(val, name);
		soln.put(sensName, SS);
	    }
	    solutions.add(soln);
	    pointer += names.size();
	}

	//next is the reply address
	setAttribute("replyAddr", tok.nextToken().trim());
	
	//now the lock
	String lockData = tok.nextToken().trim();
	//parse out the lock information
	//remove the field descriptor
	lockData = lockData.substring(3);
	StringTokenizer ST = new StringTokenizer(lockData, ",");
	l.setName(sender);
	l.setTime((new Integer(ST.nextToken())).intValue());
	l.setLID((new Integer(ST.nextToken())).intValue());
	l.setPriority(0);

	//last the timeout
	
	setAttribute("Timeout", new Integer(tok.nextToken().substring(3)));
	return solutions;
    }


    void bindSolution(Track track, Hashtable solutions, ObjEntry currObjLvl){
	/*this method takes a solution in sensor major order
	  a track object and a hashtable of sensor data
	  and updates the currrent track state with a new solution
	  as well as sends out new binding messages to enact the solution
	 */

	//one last check on the validity of this solution
	//I think that we should just return without doing any damage

	for(Iterator i = solutions.keySet().iterator();i.hasNext();){
	    if(!track.inContext((String) i.next())){
		return;
	    }
	}

	//remove agents that aren't needed any more
	Vector currSol = track.getAgents(); 
	for(int i=0;i<currSol.size();i++){
	    if(!solutions.containsKey(currSol.elementAt(i))){
		//remove the bastard
		String name = (String) currSol.elementAt(i);
		sendRB(name, new SlotSchedule()); 
		track.removeAgent(name);
	    }
	}

	Hashtable oldAssignments = track.getAssScheds();
	Enumeration e = solutions.keys();
	while(e.hasMoreElements()){
	    String SName = (String) e.nextElement();
	    SlotSchedule SS = (SlotSchedule) solutions.get(SName);
	    SlotSchedule oldSS = null;

	    if(!oldAssignments.containsKey(SName)){
		track.addAgent(SName, SS);
	    }
	    else{	      
		track.setAgentState(SName, SS);
		oldSS = (SlotSchedule) oldAssignments.get(SName);
	    }

	    //if(oldSS == null || !oldSS.equals(SS)){
		sendRB(SName, SS);
		//} 
	}
	track.getObjFunc().setCurrObj(currObjLvl);
    }

    Vector orderAlternatives(Vector solutions, Hashtable scheds, ObjEntry obj){
	/*this function is responsible for ordering the partial solutions
	  that are given in an ES message and returning a string vector 
	  that is the ordering of how good I think each of the solutions is
	  
	  I am basing my assesment on two factors
	  First:  how many conflicts will it create for me to actually
	          implement the requested solution (0 is good)
	  Second: how many non-conflicted sensors that will be 
	          changed in order to implement the solution
    
    
	  Vector (solutions)
	    |
	    elements: Hashtable (sensors)
	                 |
			 key: sensor name
			 elements: SlotSchedule


	  Hashtable (scheds)
	    |
	    key: sensor name
	    elements: SlotSchedule

	*/

	//count the number of empty and currently used schedule slots 
	//for the non-conflicting sensors
	
	int[] slotNoneCounts = new int[SlotSchedule.NUMSLOTS];
	int[] slotMyCounts = new int[SlotSchedule.NUMSLOTS];
	int[] slotConCounts = new int[SlotSchedule.NUMSLOTS];


	for(Iterator h = scheds.values().iterator();h.hasNext();){
	    SlotSchedule SS = (SlotSchedule) h.next();
	    for(int i=0;i<SlotSchedule.NUMSLOTS;i++){
		if(SS.isEmpty(i))
		    slotNoneCounts[i] += 1;
		else if(SS.hasCommitment(i,name) && SS.numCommitments(i) >1)
		    slotConCounts[i] += 1;
		else if(SS.hasCommitment(i,name))
		    slotMyCounts[i] += 1;
	    }	
	}

	//we have the non-conflicted slot counts now,
	//we have to look at the solutions and figure out
	//the conflict and change counts for each solution
	//and order them based on that

	Vector solVals = new Vector();  //keeps the solutions costs
	for(int i=0;i<solutions.size();i++){
	    Hashtable soln = (Hashtable) solutions.elementAt(i);
	    //for each slot, count the number of sensors we're
	    //given
	    int[] givenSlots = new int[SlotSchedule.NUMSLOTS];
	    for(Iterator h = soln.values().iterator();h.hasNext();){
		SlotSchedule SS = (SlotSchedule) h.next();
		for(int j=0;j<SlotSchedule.NUMSLOTS;j++){
		    if(SS.hasCommitment(j, name))
			givenSlots[j] += 1;
		}	
	    }

	    //good, we have the number of slots we're given for
	    //this solution let's figure out the change and conflict counts
	    //for each of the slots
	    Vector slotVals = new Vector();
	    for(int q =0;q<SlotSchedule.NUMSLOTS;q++) slotVals.add(new CompVal(0,0f));
	    //get the conflict counts for each of the slots
	    for(int j=0;j<SlotSchedule.NUMSLOTS;j++){
		int slotMy = slotMyCounts[j];
		int slotEmp = slotNoneCounts[j];
		int slotCon = slotConCounts[j];
		if((obj.numAgents() - scheds.size())> givenSlots[j]){
		    //this means that the other track manager had no intentions
		    //of us using this slot for our solution
		    ((CompVal) slotVals.elementAt(j)).conflicts = Integer.MAX_VALUE;
		}
		else{
		    //we can use this slot if we want

		    //conflicts is just the number of slots that we want
		    //minus the number that are freely available to us
		    //meaning the number given for the conflicting sensors
		    //+the number we can get from the non-conflicting sensors
		    ((CompVal) slotVals.elementAt(j)).conflicts = Math.max(obj.numAgents() -(givenSlots[j] + slotMyCounts[j] +slotNoneCounts[j]), 0);
		}
	    }



	    //now that we have the slot counts, we need to sort them
	    //so we can get the best obj.numSlots() slots.  Then
	    //using this ordering we can figure out the cost of adopting
	    //this solution
	    Vector slotOrder= orderByCompVal(slotVals);
	    CompVal solVal = new CompVal(0,0f);
	    for(int j =0;j<obj.numSlots();j++){
		int slotNum = ((Integer) slotOrder.elementAt(j)).intValue();
		solVal.add((CompVal) slotVals.elementAt(slotNum));
	    }
	    //add these number to the counts, because this is the number of
	    //conflicts for the solution
	    solVals.add(solVal);
	}

	//now we should order the solutions
	return orderByCompVal(solVals);
	
    }

    Vector orderByCompVal(Vector vals){
	/*this function implements a two criteria insertions sort
	  it works a bit differently though in that it returns a vector
	  with integer elements which is a sort map.  For example,
	  if you pass in the list 3,2,5,1,8 it will sort it and return
	  3,1,0,2,4 representing the ordering 1,2,3,5,8

	*/

	Vector sol = new Vector();
	int position = 0;
	for(int i=0;i<vals.size();i++){
	    for(int j = 0;j<sol.size();j++){
		int solj = ((Integer) sol.elementAt(j)).intValue();
		CompVal oldVal = (CompVal) vals.elementAt(solj);
		CompVal newVal = (CompVal) vals.elementAt(i);
		//log.log("Comparing conflicts for " + i +":" + newConfs + " to " + j+":"+oldConfs, log.LOG_INFO, FAC_PAC);
		if(oldVal.compareTo(newVal) < 0){
		    position = j;
		    break;
		}
		else{
		    position++;
		}
	    }
	    sol.insertElementAt(new Integer(i), position);
	    position =0;
	}
	return sol;
     
    }

    private void sendRB(String who, SlotSchedule SS){
	String msg = "RB " + SS.getCommitmentsInt(name);
	Message m = new Message(name, who, msg);
	log.log("Sending rebind " + m.toString(),log.LOG_INFO);
	me.sendMessage(m);
    }

} // SPAMR


class CompVal implements Comparable{
    public int conflicts;
    public float score;
    
    public void add(CompVal o1){
	this.conflicts += o1.conflicts;
	this.score += o1.score;
    }

    public CompVal(int conf, float sc){
	conflicts = conf;
	score = sc;
    }

    public int compareTo(Object o1){
	CompVal first = this;
	CompVal second = (CompVal) o1;
	
	
	if(first.conflicts > second.conflicts){
	    return -1;
	}
	else if(first.conflicts < second.conflicts){
	    return 1;
	}
	else{
	    if(first.score > second.score){
		return 1;
	    }
	    else if(first.score < second.score){
		return -1;
	    }
	    else{
		return 0;
	    }
	
	}
    }
    public boolean equals(Object obj){
	CompVal other = (CompVal) obj;
	return (this.conflicts == other.conflicts && this.score == other.score);
    }

}
