package farm.ants.agent;
import farm.simulator.*;
import farm.metagent.*;
import java.util.*;

/**
 * A simple representation of a sensor agent.  Just maintains a schedule
 * and informs track managers that are registered with it, when changes occur.
 * <p>
 * Created: Mon Jul 15 10:26:42 2002
 *
 * @author Roger Mailler
 */
public class SensorAgent extends Agent {
    SlotSchedule sched = new SlotSchedule();

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agents name
     */
    public SensorAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);
    }

    /**
     * Called by the metagent to grant run time to this agent.
     * The SensorAgent simple pulls any incoming messages
     * off of its incoming message queue and if they are a rebing request
     * changes it local schedule and sends update messages to registered 
     * track managers.
     */
    public void pulse() {
	super.pulse();
        checkRun();
	boolean change = false;
	//log.log("Sensor got pulsed",log.LOG_INFO);

        for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
            //log.log("Got message: " + msg, Log.LOG_INFO);
            checkRun();
	    //change the schedule according to the RB message
	    if(msg.getData().startsWith("RB")){
		int val= Integer.parseInt(msg.getData().substring(3));
		sched.addInt(val,msg.getSource());
		change = true;
	    }
        }
	if(change){
	    Vector mans = sched.getTracks();
	    for(int i =0;i<mans.size();i++){
		sendMessage(new Message(getAgentName(),(String) mans.elementAt(i) ,"RS " + sched.toPiggyBack()));
	    }
	    //update the server with my current schedule
	    setProperty(getAgentName() + ":sched", sched.toPiggyBack());
	    change = false;
	}
	endRun();
    }
} // SensorAgent
