package farm.ants.plugins;

import java.rmi.*;
import java.rmi.server.*;

import farm.simulator.*;

import java.util.*;
import farm.ants.agent.*;
import farm.generic.plugins.*;
import farm.generic.plugins.drivers.*;
/**
 */
public class AntsAllocationAnalysis extends FarmPlugIn {

    // Constants
    public static final String PROP_TOTALSENSORS = "TotalSensors";
    public static final String PROP_TOTALTARGETS = "TotalTargets";
    public static final String PROP_TOTALMESSAGESSENT = "TotalMessagesSent";
    public static final String PROP_TOTALMESSAGESRECEIVED = "TotalMessagesReceived";
    public static final String PROP_TOTALBYTESSENT = "TotalBytesSent";
    public static final String PROP_TOTALBYTESRECEIVED = "TotalBytesReceived";
    public static final String PROP_TOTALUTILITY = "TotalUtility";
    public static final String PROP_POSSIBLEUTILITY = "PossibleUtility";
    public static final String PROP_POSSIBLETRACKS = "PossibleTracks";
    public static final String PROP_TOTALTRACKS = "TotalTracks";
    public static final String PROP_POSSIBLECONFLICTS = "PossibleConflicts";
    public static final String PROP_TOTALCONFLICTS = "TotalConflicts";



    /**
     * Constructor
     * @param address The address of the remote simulator to eonnect to
     */
    public AntsAllocationAnalysis(String []args)throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Ants Allocation Analysis"));
    }

    /**
     * This is called at the end of a pulse.
     */
    public void pulseEnd() {

        System.err.println("Pulse start");
        try {
            //LocalFunction function = new AllocationAnalysis();            
            runLocalFunction(new AllocationAnalysis());
            //function.run(simulator);

        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
        System.err.println("Pulse end");
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmPlugIn plugin = new AntsAllocationAnalysis(args);
            plugin.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}

class AllocationAnalysis extends AnalysisFunction {



    int slotCount = 3;

    public Object run(PropertyOwner owner) {
        int sensors = 0,
            targets = 0,
            msgsent = 0,
            msgrecv = 0,
            bytesent = 0,
            byterecv = 0;

        try {
	    int time = ((Integer) owner.getProperty(Simulator.P_TIME)).intValue();
	    
            // Count sensors
	    ArrayList mans = new ArrayList();
            Collection list = owner.findProperties(":loc");
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                if (name.startsWith("SA"))
                    sensors++;
                else if (name.startsWith("M")){
                    targets++;
		    mans.add(name.substring(0, name.indexOf(":")));
		}
            }

            //owner.setProperty(AntsAllocationAnalysis.PROP_TOTALSENSORS, new Integer(sensors));
            //owner.setProperty(AntsAllocationAnalysis.PROP_TOTALTARGETS, new Integer(targets));

            // Track messaging
            list = owner.findProperties(":MessagesSent");
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Number num = (Number)owner.getProperty(name);
                msgsent += num.intValue();
            }
            owner.setProperty(AntsAllocationAnalysis.PROP_TOTALMESSAGESSENT, new Integer(msgsent));
            list = owner.findProperties(":MessagesReceived");
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Number num = (Number)owner.getProperty(name);
                msgrecv += num.intValue();
            }
            owner.setProperty(AntsAllocationAnalysis.PROP_TOTALMESSAGESRECEIVED, new Integer(msgrecv));

            list = owner.findProperties(":BytesSent");
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Number num = (Number)owner.getProperty(name);
                bytesent += num.intValue();
            }
            owner.setProperty(AntsAllocationAnalysis.PROP_TOTALBYTESSENT, new Integer(bytesent));
            list = owner.findProperties(":BytesReceived");
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                Number num = (Number)owner.getProperty(name);
                byterecv += num.intValue();
            }
            owner.setProperty(AntsAllocationAnalysis.PROP_TOTALBYTESRECEIVED, new Integer(byterecv));
	    
	    //figure out maximum possible allocation
            list = owner.findProperties(":res");
	    Hashtable res = new Hashtable();
	    float total = 0.0f;
	    int totTrack = 0;
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
		String prop = (String)owner.getProperty(name);
                if (prop != null) {
                    StringTokenizer tok = new StringTokenizer(prop,",");
                    res.put(name.substring(0, name.indexOf(":")), owner.getProperty(name));
                    if(tok.countTokens() >2){	
                        total += 3.0f * (float) Math.sqrt((Math.min(tok.countTokens(),4))-2);
                        totTrack++;
                    }
                }
            }
	    owner.setProperty(AntsAllocationAnalysis.PROP_POSSIBLEUTILITY, new Float(total));
	    owner.setProperty(AntsAllocationAnalysis.PROP_POSSIBLETRACKS, new Integer(totTrack));
	    owner.setProperty(AntsAllocationAnalysis.PROP_POSSIBLECONFLICTS, new Integer(calculatePossConflicts(res)));


	    //get the sensors schedules
            list = owner.findProperties(":sched");
	    Hashtable scheds = new Hashtable();
            for (Iterator i = list.iterator(); i.hasNext(); ) {
                String name = (String)i.next();
                String prop = (String)owner.getProperty(name);
                if (prop != null) {
                    SlotSchedule SS = new SlotSchedule();
                    SS.fromPiggyBack(prop);
                    scheds.put(name.substring(0, name.indexOf(":")), SS);
                }
            }

	    float tempFloat = calculateUtility(scheds, mans, res, totTrack);
	    owner.setProperty(AntsAllocationAnalysis.PROP_TOTALUTILITY, new Float(tempFloat));

	    /*Float oFloat = (Float) owner.getProperty(AntsOptimalAllocation.PROP_OPTIMALUTILITY);
	    
	      if(oFloat != null && tempFloat > oFloat.floatValue() && calculateTotConflicts(scheds, res) == 0){
	      
	      System.out.println("!!!!!!!ERROR!!!!!!!!: " + scheds.toString());
	    }*/


	    owner.setProperty(AntsAllocationAnalysis.PROP_TOTALTRACKS, new Integer(calculateTotTracks(scheds, mans, res)));
	    int tconflicts = calculateTotConflicts(scheds, res);
	    owner.setProperty(AntsAllocationAnalysis.PROP_TOTALCONFLICTS, new Integer(tconflicts));
	    
	    if(msgrecv == 0 && msgsent == 0 && tconflicts == 0 && list.size()>0 ){
		owner.setProperty(StopDriver.STOPSIMULATION, new Boolean(true));  
	    }
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }

        return null;
    }

    private int calculatePossConflicts(Hashtable res){
	ArrayList counts = new ArrayList();
	int total = 0;
	for(Iterator i = res.values().iterator();i.hasNext();){
	    StringTokenizer ST = new StringTokenizer((String) i.next(), ",");
	    int thisTotal = 0;
	    int checked = 0;
	    while(ST.hasMoreTokens()){
		String dat = ST.nextToken();
		String name = dat.substring(0, dat.indexOf(" "));
		if(counts.contains(name)){
		    thisTotal++;
		}
		else{
		    counts.add(name);
		}
		checked++;
		if(checked == 4) break;
	    }
	    total += Math.min(thisTotal,4);
	}
	return 3*total;
    }
    
    private int calculateTotConflicts(Hashtable scheds, Hashtable res){
	int conflict = 0;

	//get a list of the required sensors
	ArrayList resVect = new ArrayList();
	for(Iterator i = res.values().iterator();i.hasNext();){
	    String resString = (String) i.next();  //the resources for a target
	    StringTokenizer ST = new StringTokenizer(resString, ",");
	    while(ST.hasMoreTokens()){
		String dat = ST.nextToken();
		String sens = dat.substring(0,dat.indexOf(" "));
		if(!resVect.contains(sens)){
		    resVect.add(sens);
		}		
	    }
	}
	for(Iterator i = resVect.iterator();i.hasNext();){
	    SlotSchedule SS = (SlotSchedule) scheds.get(i.next());
	    if(SS == null) continue;
	    for(int j=0;j<slotCount;j++){
		if(SS.numCommitments(j) > 1)
		    conflict += SS.numCommitments(j) -1;
	    }
	    
	}
	return conflict;
    }

    
    private int calculateTotTracks(Hashtable scheds, ArrayList mans, Hashtable res){
	int totalTracks = 0;
	for(Iterator i = mans.iterator();i.hasNext();){
	    String name = (String) i.next();
	    float bestProb = 0.0f;
	    String resString = (String) res.get(name);
	    if(resString == null) continue;  //no resources to track
	    StringTokenizer ST = new StringTokenizer(resString, ",");
	    Vector resVect = new Vector();
	    if(ST.countTokens() < 3) continue; //not enough to track
	    while(ST.hasMoreTokens()) {
		String dat = ST.nextToken();
		String sens = dat.substring(0,dat.indexOf(" "));
		resVect.add(sens);
	    }
	    for(int j=0;j<slotCount;j++){
		float thisProb = calcSlotProb(scheds, resVect, name, j);
		if(thisProb > bestProb) bestProb = thisProb;
	    }
	    if(bestProb == 1)
		totalTracks ++;
	}
	return totalTracks;
    }


    private float calculateUtility(Hashtable scheds, ArrayList mans, Hashtable res, int possTracks){
	float totalUtil = 0.0f;
	for(Iterator i = mans.iterator();i.hasNext();){
	    String name = (String) i.next();
	    float manUtil = 0.0f;
	    String resString = (String) res.get(name);
	    if(resString == null) continue;
	    StringTokenizer ST = new StringTokenizer(resString, ",");
	    Vector resVect = new Vector();
	    while(ST.hasMoreTokens()) {
		String dat = ST.nextToken();
		String sens = dat.substring(0,dat.indexOf(" "));
		resVect.add(sens);
	    }
	    for(int j=0;j<slotCount;j++){
		manUtil += calcSlotUtil(scheds, resVect, name, j);
	    }
	    totalUtil += manUtil;
	}
	int totTracks = calculateTotTracks(scheds, mans, res);
	float ret = totalUtil - 2.0f *(possTracks - totTracks);
	
	System.out.println("totalUtil " + totalUtil + " possTracks " + possTracks + " totTracks " + totTracks + " return " + ret);

	return ret;
    }

    private float calcSlotUtil(Hashtable sensors, Vector res, String manager, int slot){
	float slotUtil = 0.0f;
	Vector sens = new Vector();

	//get a list of the sensors that have commitment
	//for this track manager in this slot and are part of the requirment
	for(Iterator i = sensors.keySet().iterator();i.hasNext();){
	    String name = (String) i.next();
	    SlotSchedule s = (SlotSchedule) sensors.get(name);
	    if(s.hasCommitment(slot, manager) && res.contains(name)){
		sens.add(name);
	    }
	}

	//now generate every combination of size 3 to sens.size()
	for(int i=3;i<=sens.size();i++){
	    Vector combs = Choose(sens, i);
	    //go through every combination and calculate the
	    //probability that it will occur

	    for(int j = 0;j<combs.size();j++){
		slotUtil += combProb((Vector) combs.elementAt(j), sens, sensors, manager, slot) * Math.sqrt(i-2);
	    }
	}	
	return slotUtil;
    }


    private float calcSlotProb(Hashtable sensors, Vector res, String manager, int slot){
	float slotProb = 0.0f;
	Vector sens = new Vector();

	//get a list of the sensors that have commitment
	//for this track manager in this slot and are part of the requirment
	for(Iterator i = sensors.keySet().iterator();i.hasNext();){
	    String name = (String) i.next();
	    SlotSchedule s = (SlotSchedule) sensors.get(name);
	    if(s.hasCommitment(slot, manager) && res.contains(name)){
		sens.add(name);
	    }
	}

	//now generate every combination of size 3 to sens.size()
	for(int i=3;i<=sens.size();i++){
	    Vector combs = Choose(sens, i);
	    //go through every combination and calculate the
	    //probability that it will occur

	    for(int j = 0;j<combs.size();j++){
		slotProb += combProb((Vector) combs.elementAt(j), sens, sensors, manager, slot);
	    }
	}	
	return slotProb;
    }




    private float combProb(Vector comb, Vector sens, Hashtable sensors, String manager, int slot){
	float prob = 1.0f;

	//combination probability is the probability that a sensor not in the
	//combination will not actually measure * probability that sensors in
	//the list actually measure


	for(int i =0;i<sens.size();i++){
	    String sName = (String) sens.elementAt(i);
	    if(!comb.contains(sName)){
		SlotSchedule s = (SlotSchedule) sensors.get(sName);
		prob *= 1.0f - (1.0f/(float) s.numCommitments(slot));
	    }
	}
	for(int i=0;i<comb.size();i++){
	    String sName = (String) comb.elementAt(i);
	    SlotSchedule s = (SlotSchedule) sensors.get(sName);
	    prob *= 1.0f/(float) s.numCommitments(slot);
	}
	return prob;
    }

    private Vector Choose(Vector elements, int r){
	//this function takes a vector of n elements and returns a vector
	//that contains the nCr possible combinations (vectors) of the elements
	
	Vector retData = new Vector();
	if(r>0){
	    nCr(r, 0, new Vector(), 0, elements.size(), elements, retData);
	}
	return retData;
    }
    
    
    private void nCr(int maxdepth, int currdepth, Vector solution, int first, 
	     int last, Vector elements, Vector Sets){
	
	for(int i = first;i<= last-maxdepth+currdepth;i++){
	    solution.add(elements.elementAt(i));
	    if(currdepth < maxdepth-1){
		nCr(maxdepth, currdepth+1, solution, i+1, last, elements, Sets);
		solution.remove(currdepth);
	    }
	    else{
		//Operate on solution
		Sets.add(solution.clone());
		solution.remove(currdepth);
	    }
	}
    }
}
