package farm.ants.plugins;
/**
 * FarmDisplayFrame.java
 *
 *
 * Created: Wed Jul 10 14:05:04 2002
 *
 * @author Roger Mailler
 * @version
 */
import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.*;
public class FarmDisplayFrame extends JFrame
    implements ActionListener {
    
    String title = "ANT Farm World Display";
    JPanel JP = new JPanel();
    JScrollPane JSP;
    FarmWorldPanel topPanel = new FarmWorldPanel(this);
    JPanel bottomPanel = new JPanel();
    JMenu menuFile;
    JMenuBar menubar;
    JMenuItem miAbout, miQuit;
    HashMap targets = null;
    ArrayList sensors = null;

    public FarmDisplayFrame() {
	setTitle(title);
	getContentPane().setLayout(new BorderLayout());
	addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
		    
		    System.exit(0);
		}
	    });

        JSP = new JScrollPane();
        JSP.setHorizontalScrollBarPolicy(JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
        JSP.setVerticalScrollBarPolicy(JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED);
        JSP.getViewport().add(topPanel);

        getContentPane().add(JP);
	JP.setLayout(new BorderLayout());
	JP.add(JSP, BorderLayout.CENTER);
	JP.add(bottomPanel, BorderLayout.SOUTH);
	
	buildMenus();
    }
    public void actionPerformed(ActionEvent e) {

	if (e.getSource() == miAbout) {
	    JOptionPane.showMessageDialog(null,
					  "ANT Farm Display Plugin\n\nby Roger Mailler\n\nMulti-Agent Systems Lab\nUniversity of Massachusetts\nhttp://mas.cs.umass.edu/",
					  "About the Display",
					  JOptionPane.INFORMATION_MESSAGE);
	    
	} 
	else if (e.getSource() == miQuit)
	    System.exit(0);

    }


    private void buildMenus(){
 	
	menubar = new JMenuBar();

	menuFile = new JMenu("File");
	this.setJMenuBar(menubar);
	menubar.add(menuFile);
	miAbout      = new JMenuItem("About");
	miQuit       = new JMenuItem("Quit");
	
	menuFile.add(miAbout);
        menuFile.addSeparator();
	menuFile.add(miQuit);
	miAbout.addActionListener(this);
	miQuit.addActionListener(this);

        JMenu menuPrefs = new JMenu("Preferences");
        menubar.add(menuPrefs);

        JCheckBoxMenuItem miNegot = new JCheckBoxMenuItem("Show Negotiations", topPanel.getShowNegotiations());
        miNegot.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JCheckBoxMenuItem item = (JCheckBoxMenuItem)e.getSource();
                    item.setSelected(item.isSelected());
                    topPanel.setShowNegotiations(item.isSelected());
                }
            });
        menuPrefs.add(miNegot);

        JCheckBoxMenuItem miTrack = new JCheckBoxMenuItem("Show Tracking", topPanel.getShowTracking());
        miTrack.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JCheckBoxMenuItem item = (JCheckBoxMenuItem)e.getSource();
                    item.setSelected(item.isSelected());
                    topPanel.setShowTracking(item.isSelected());
                }
            });
        menuPrefs.add(miTrack);

        JCheckBoxMenuItem miTargText = new JCheckBoxMenuItem("Show Target Text", topPanel.getShowTargetText());
        miTargText.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JCheckBoxMenuItem item = (JCheckBoxMenuItem)e.getSource();
                    item.setSelected(item.isSelected());
                    topPanel.setShowTargetText(item.isSelected());
                }
            });
        menuPrefs.add(miTargText);

        JCheckBoxMenuItem miSensText = new JCheckBoxMenuItem("Show Sensor Text", topPanel.getShowSensorText());
        miSensText.addActionListener(new ActionListener() {
                public void actionPerformed(ActionEvent e) {
                    JCheckBoxMenuItem item = (JCheckBoxMenuItem)e.getSource();
                    topPanel.setShowSensorText(!topPanel.getShowSensorText());
                    item.setSelected(topPanel.getShowSensorText());
                }
            });
        menuPrefs.add(miSensText);
    }

    public void updateDisplay(){
	topPanel.updateDisplay();
    }

    public void setSensors(ArrayList sens){
	sensors = sens;
	topPanel.updateDisplay();
    }
    public void setTargets(HashMap targs){
	targets = targs;
    }

    public ArrayList getSensors(){
	return sensors;
    }
    public HashMap getTargets(){
	return targets;
    }

    public void setHeight(int h){
	topPanel.setEnvHeight(h);
    }
    public void setWidth(int w){
	topPanel.setEnvWidth(w);
    }

    public JScrollPane getJSP(){
	return JSP;
    }


} // FarmDisplayFrame
