package farm.ants.plugins;
import farm.simulator.*;
import java.util.*;
import java.rmi.*;
import java.awt.geom.*;
import farm.ants.agent.SensorData;

/**
 * This class describes the movement model and location of a particular target.
 * <p>
 * Created: Tue Jul  9 14:16:09 2002
 *
 * @author Roger Mailler
 */
public class TargetModel extends ObjectModel implements Comparable{
    Random rnd = new Random();

    /**
     * Constructor
     */
    public TargetModel(FarmPlugIn FPI, String n){
	super(FPI, n);
    }

    /**
     * Constructor
     */
    public TargetModel(FarmPlugIn FPI, String n, float X, float Y, 
		       float speed){
	super(FPI, n, X, Y);
	setMaxSpeed(speed);
    }

    float maxSpeed;
    
    /**
       * Get the value of maxSpeed (feet/sec).
       * @return Value of maxSpeed (feet/sec).
       */
    public float getMaxSpeed() {return maxSpeed;}
    
    /**
       * Set the value of maxSpeed (feet/sec) and 
       * sets the targets actual velocity to be some speed
       * between 0 and maxSpeed with a random vector
       * @param v  Value to assign to maxSpeed (feet/sec).
       */
    public void setMaxSpeed(float  v) {
	this.maxSpeed = v;
	float newSpeed = rnd.nextFloat()*maxSpeed;
	float vx = rnd.nextFloat() * newSpeed;
	float vy = (float) Math.sqrt((newSpeed*newSpeed) - (vx * vx));
	setVX((rnd.nextBoolean()?1.0f:-1.0f)*vx);
	setVY((rnd.nextBoolean()?1.0f:-1.0f)*vy);
    }
    
    float vX;
    
    /**
       * Get the value of vX.
       * @return Value of vX.
       */
    public float getVX() {return vX;}
    
    /**
       * Set the value of vX.
       * @param v  Value to assign to vX.
       */
    public void setVX(float  v) {this.vX = v;}
    

    float vY;
    
    /**
       * Get the value of vY.
       * @return Value of vY.
       */
    public float getVY() {return vY;}
    
    /**
       * Set the value of vY.
       * @param v  Value to assign to vY.
       */
    public void setVY(float  v) {this.vY = v;}
    
    /**
       * Move the target.  Will reverse taget direction
       * upon hitting a wall, and will randomly reset the
       * targets velocity vector
       * @param time milliseconds since last move.
       */
    public void move(int time){
	Point2D.Float ptf = getPoint(false);

	if(rnd.nextFloat() < 0.001) setMaxSpeed(getMaxSpeed());

	float newX = (vX*(float)time/1000f) +(float) ptf.getX();
	float newY = (vY*(float)time/1000f) +(float) ptf.getY();
	
	if(newX >= getRoomWidth() || newX <= 0){
	    vX = -vX;
	    newX = (vX*(float)time/1000f) +(float) ptf.getX();
	}
	if(newY >= getRoomHeight() || newY <= 0){
	    vY = -vY;
	    newY = (vY*(float)time/1000f) +(float) ptf.getY();
	}
	ptf.setLocation(newX, newY);
	setPoint(ptf);
    }


    ArrayList sensors = new ArrayList();

    /**
       * Sets the list of sensors that can see a target.  Updates
       * the server if a change has occured
       * 
       * @param the set of sensors that can see this target (Vector of SensorData)
       */
    public void setSensors(ArrayList a) {
	//do we need to change the list?
	int j = sensors.size();
	boolean update = false;
	String uString = "";
	for(int i = 0;i<a.size();i++){
	    SensorData mdl = (SensorData) a.get(i);
	    uString = uString + mdl.toString() + ((i<a.size()-1)?",":"");
	    if(!sensors.contains(mdl))
		update = true;
	    else 
		j--;
	}
	if(update || j>0) fpi.setProperty(name + ":res", uString); 
	sensors = a;
    }
    
    int roomWidth = 100;
    
    /**
       * Get the value of roomWidth.
       * @return Value of roomWidth.
       */
    public int getRoomWidth() {return roomWidth;}
    
    /**
       * Set the value of roomWidth.
       * @param v  Value to assign to roomWidth.
       */
    public void setRoomWidth(int  v) {this.roomWidth = v;}
    
    int roomHeight = 100;
    
    /**
       * Get the value of roomHeight.
       * @return Value of roomHeight.
       */
    public int getRoomHeight() {return roomHeight;}
    
    /**
       * Set the value of roomHeight.
       * @param v  Value to assign to roomHeight.
       */
    public void setRoomHeight(int  v) {this.roomHeight = v;}
    
    public String getState(){return (String) fpi.getProperty(name + ":state");}


} // TargetModel
