package farm.bandit.agent;

import farm.simulator.*;
import farm.metagent.*;

import java.util.*;
import utilities.Log;

/**
 * This agent implements the eGreedy learning algorithm,
 * according to (Sutton, Barto 1998) p.28.
 */
public class eGreedyAgent extends BanditAgent {
    public static String NAME = "greedy";
    protected double e = 0.1;
    protected double q0 = 5;
    protected double q[];

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     */
    public eGreedyAgent(FarmMetAgent metagent, String name) {
	super(metagent, name);
    }

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param name The agent's name
     * @param e The epsilon value to use
     * @param q0 The initial q value to assign to each arm
     */
    public eGreedyAgent(FarmMetAgent metagent, String name, double e, double q0) {
        this(metagent, name);

        this.e = e;
        this.q0 = q0;
    }

    /**
     * Initialize the state vector
     */
    public void begin() {
        super.begin();

	// Set the initial q values
        q = new double[getNumArms()];
        for (int i = 0; i < q.length; i++) {
            q[i] = q0;
        }
    }

    /**
     * Run a single play of the bandit.
     */
    public void pulse() {
        super.pulse();
        int arm;

        // First decide if we are greedy
        boolean greedy = random.nextFloat() >= e;

	// Then pick an arm
        if (greedy) {
            arm = random.nextInt(q.length);
            for (int i = 0; i < q.length; i++) {
                if (q[i] > q[arm]) {
                    arm = i;
                }
            }
        } else {
            arm = random.nextInt(q.length);
        }

        // Pull the arm
        double reward = pullArm(arm);

	// Incrementally update the q value
        q[arm] = q[arm] + (1d/((double)getNumPulls(arm)+1)) * (reward - q[arm]);

        //System.err.println(getName() + " Greedy:" + greedy + " Arm:" + arm + " Reward:" + reward + " Optimal:" + isOptimalArm(arm) + " q:" + q[arm]);
    }
}
