package farm.bandit.plugins;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;

import farm.simulator.*;
import farm.metagent.*;
import farm.bandit.agent.*;

import utilities.Log;

/**
 * This component is responsible for spawning and managing
 * a collection of learning agents.  Its assignment will
 * have been previously specified by the BanditDriver. 
 * @see #createAllAgents
 * @see BanditDriver
 */
public class BanditMetAgent extends FarmMetAgent {
    
    // Constants
    protected String lastdescription = null;

    /**
     * Constructor
     * @param simulator The remote simulator object to use
     * @param name The name of the metagent
     */
    public BanditMetAgent(String []args) throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "Bandit MetAgent"));
    }

    /**
     * This function actually creates the agents.
     */
    public void createAllAgents() {

        // First flush existing agents
        stopAllAgents();

        // Then create the new ones
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        lastdescription = (String)getProperty(prop);
        if (lastdescription != null) {
            StringTokenizer tok = new StringTokenizer(lastdescription);
            while (tok.countTokens() > 1) {
                StringTokenizer type = new StringTokenizer(tok.nextToken(), ",");
                String range = tok.nextToken();

                StringTokenizer rtok = new StringTokenizer(range, "-");
                int start = Integer.parseInt(rtok.nextToken());
                int finish = Integer.parseInt(rtok.nextToken());
                
                String algorithm = type.nextToken();
                double args[] = new double[type.countTokens()];
                for (int i = 0; type.hasMoreTokens(); i++) {
                    args[i] = Double.parseDouble(type.nextToken());
                }

                log.log("Creating " + algorithm + " " + start + "-" + finish, Log.LOG_INFO);
                for (int i = start; i <= finish; i++) {
                    if (algorithm.equals(eGreedyAgent.NAME)) {
                        addAgent(new eGreedyAgent(this, "G" + args[0] + "-" + args[1] + ":" + i, args[0], args[1]));

                    } else if (algorithm.equals(PursuitAgent.NAME)) {
                        addAgent(new PursuitAgent(this, "P" + args[0] + ":" + i, args[0]));

                    } else if (algorithm.equals(ReinforcementComparisonAgent.NAME)) {
                        addAgent(new ReinforcementComparisonAgent(this, "RC" + args[0] + ":" + i, args[0]));

                    } else if (algorithm.equals(ReinforceAgent.NAME)) {
                        addAgent(new ReinforceAgent(this, "R" + args[0] + ":" + i, args[0], args[1], args[2]));

                    } else {
                        log.log("Unknown type id " + type, Log.LOG_WARNING);
                    }
                }
            }
        } else {
            log.log("Could not find agent description property " + prop, Log.LOG_WARNING);
        }
    }

    /**
     */
    protected void pulseBegin() {

        super.pulseBegin();

        // Look for a new description
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        String agentsdesc = (String)getProperty(prop);
        if (agentsdesc != null) {
            if (lastdescription == null || !lastdescription.equals(agentsdesc)) {
                createAllAgents();
            }
        }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmMetAgent agent = new BanditMetAgent(args);
            agent.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
