package farm.graphcolor.agent;

import farm.simulator.*;
import farm.metagent.*;
import farm.graphcolor.plugins.drivers.*;

import java.util.*;
import utilities.Log;

/**
 */
public class DBOAgent extends Agent {
    NodeModel model;
    Hashtable view = new Hashtable();
    Hashtable neighbors = new Hashtable();
    Random random = new Random();
    private static int MODE_STARTUP = 0;
    private static int MODE_WAIT = 1;
    private static int MODE_IMPROVE = 2;
    int currMode = MODE_STARTUP;
    int my_termination_counter = 0;
    boolean quasi_local_minimum = false;
    boolean can_move = false;
    boolean consistent = false;
    int new_value = -1;
    Vector message_queue = new Vector();
    int counter = 0;
    int my_improve = 0;
    int max_distance;

    /**
     * Constructor
     * @param metagent A handle to the metagent this agent belongs to
     * @param m The NodeModel assigned to this agent
     */
    public DBOAgent(FarmMetAgent metagent, NodeModel m) {
	super(metagent, m.getName());

        model = m;
	//randomly pick a color
     
	//model.setColor(random.nextInt(model.getPossibleColors()));
	for (Iterator i = model.getNeighbors(); i.hasNext();) {
	    neighbors.put((String) i.next(), new Integer(1));
	}

	max_distance = ((Integer) getProperty(GraphColorDriver.CLA_NODES)).intValue()/3;

    }

    /**
     */
    public void pulse() {
	super.pulse();
	if(currMode == MODE_STARTUP){
	    send_ok();
	    currMode = MODE_WAIT;
	}

        // Respond to messages
        for (Message msg = getFirstIncomingMessage(); msg != null; msg = getFirstIncomingMessage()) {
            checkRun();
	    if (msg.getData().startsWith("ok? ") && currMode == MODE_WAIT) {
		processOK(msg);
            } else if (msg.getData().startsWith("improve ") && currMode == MODE_IMPROVE) {
		processImprove(msg);
            }
	    else{
		//queue the message
		message_queue.add(msg);
	    }
        }
	endRun();
    }

    private void processOK(Message msg){
	counter ++;
	DBOAgentView v = new DBOAgentView(msg.getData().substring(4));
	view.put(v.getName(), v);
	//log.log("Got an ok. Need " + (model.numNeighbors() - counter) + " more", log.LOG_INFO);
	if(counter == model.numNeighbors()){
	    send_improve();
	    counter = 0;
	    currMode = MODE_IMPROVE;
	    for(Iterator i = message_queue.iterator();i.hasNext();){
		processImprove((Message) i.next());
		i.remove();
	    }
	}
    }


    private void processImprove(Message msg){
	StringTokenizer tok = new StringTokenizer(msg.getData().substring(8));
	String who = msg.getSource();
	int improve = Integer.parseInt(tok.nextToken());
	int eval = Integer.parseInt(tok.nextToken());
	int termination_counter = Integer.parseInt(tok.nextToken());
	
	counter ++;
	my_termination_counter = Math.min(termination_counter, my_termination_counter);
	if(improve > my_improve){
	    can_move = false;
	    quasi_local_minimum = false;
	}
	if(improve == my_improve && isHigherPriority(who)){
	    can_move = false;
	}
	if(eval >0) consistent = false;
	if(counter == model.numNeighbors()){
	    send_ok();
	    counter =0;
	    view.clear();
	    currMode = MODE_WAIT;
	    for(Iterator i = message_queue.iterator();i.hasNext();){
		processOK((Message) i.next());
		i.remove();
	    }
	}
    }
  
    /**
     * Determines if an agent is more important than me
     */
    public boolean isHigherPriority(String n) {
	int p = Integer.parseInt(n.substring(2));
	int m = Integer.parseInt(model.getName().substring(2));
	//if (p > m)
	    //log.log(model.getName() + " is implicitly lower than " + n, Log.LOG_INFO);
	return (p > m);
    }

    private void send_improve(){

	int colors[] = new int[model.getPossibleColors()];
        int mycolor = model.getColor();

        // Rate the colors
        for (Iterator i = model.getNeighbors(); i.hasNext(); ) {
            DBOAgentView neighbor = (DBOAgentView)view.get((String)i.next());
	    int constraint_val =  ((Integer) neighbors.get(neighbor.getName())).intValue();
	    colors[neighbor.getColor()] += constraint_val;
        }
	
	int current_eval = colors[mycolor];
	my_improve = 0;
        new_value = mycolor;
        for (int i = 0; i < colors.length; i++) {
            // If one is clearly better, choose it
            if (colors[i] < colors[new_value]) {
                new_value = i;
		my_improve = current_eval - colors[new_value];
		
            } else if (colors[i] == colors[new_value] &&
                       random.nextInt(model.getPossibleColors()) == 0) {
                new_value = i;
            }
        }

	if(current_eval == 0){
	    consistent = true;
	}
	else{
	    consistent = false;
	    my_termination_counter = 0;
	}

	if(my_improve > 0){
	    can_move = true;
	    quasi_local_minimum = false;
	}
	else{
	    can_move = false;
	    quasi_local_minimum = true;
	}
    
    	//send an ok message to all neighbors
	for (Iterator i = model.getNeighbors(); i.hasNext(); ) {
	    sendMessage(new Message(getAgentName(), (String)i.next(), "improve " + my_improve + " " + current_eval + " " + my_termination_counter));
	}
    
    }

    private void send_ok(){
	if(consistent){
	    my_termination_counter++;
	    if(my_termination_counter == max_distance){
		//solution found send messages
	    }
	}
	if(quasi_local_minimum){
	    //increase weigths of constraint violations
	    for (Iterator i = model.getNeighbors(); i.hasNext(); ) {
		DBOAgentView neighbor = (DBOAgentView)view.get((String)i.next());
		int constraint_val =  ((Integer) neighbors.get(neighbor.getName())).intValue();
		if(neighbor.getColor() == new_value)
		    neighbors.put(neighbor.getName(), new Integer(constraint_val+1));
	    }
	}
	if(can_move){
	    //change my color to the new one
	    model.setColor(new_value);
	}
	//send an ok message to all neighbors
	for (Iterator i = model.getNeighbors(); i.hasNext(); ) {
	    sendMessage(new Message(getAgentName(), (String)i.next(), "ok? " + new DBOAgentView(model).toString()));
	}
    }
}

class DBOAgentView {
    String name;
    int color;

    public DBOAgentView(String n, int c) {
        name = n;
        color = c;
    }
    public DBOAgentView(String s) {
        StringTokenizer tok = new StringTokenizer(s);
        name = tok.nextToken();
        color = Integer.parseInt(tok.nextToken());
    }
    public DBOAgentView(NodeModel n) {
        this(n.getName(), n.getColor());
    }

    public String getName() { return name; }
    public int getColor() { return color; }

    public boolean equals(Object o) {
        DBOAgentView v = (DBOAgentView)o;
        if (v.getColor() == color &&
            v.getName().equals(name))
            return true;
        return false;
    }

    public String toString() {
        return name + " " + color;
    }
}
