package farm.sat.plugins;

import java.rmi.*;
import java.rmi.server.*;
import java.util.*;

import farm.simulator.*;
import farm.metagent.*;
import farm.sat.agent.*;
import farm.sat.plugins.drivers.*;

import utilities.Log;

/**
 * @see #createAllAgents
 */
public class SATMetAgent extends FarmMetAgent {
    
    // Constants
    protected String lastdescription = null;

    /**
     * Constructor
     * @param simulator The remote simulator object to use
     * @param name The name of the metagent
     */
    public SATMetAgent(String []args) throws RemoteException {
        super(args);

        // Command line args
        setPlugInName(getCommandLineArgument(args, CLA_NAME, "SAT MetAgent"));
    }

    /**
     * Creates the graph color node agents.  This is done by looking
     * for a specially formatted string at the property
     * "getMetAgentName() + AGENTS_DESCRIPTION".  The string should be of
     * the format "Node 0-25" - a space delimited list of token
     * pairs.  The first token in each pair (e.g. Node) indicates
     * the type of agent to create, while the second specifies the numbers
     * of agents to create (inclusive).
     */
    public void createAllAgents() {

        // First flush existing agents
        stopAllAgents();

        // Then create the new ones
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        lastdescription = (String)getProperty(prop);
        if (lastdescription != null) {
            StringTokenizer tok = new StringTokenizer(lastdescription);
            while (tok.countTokens() > 1) {
                String type = tok.nextToken();
                String range = tok.nextToken();

                int typeid = -1;
                if (type.equalsIgnoreCase("random")) {
                    typeid = 0;
                } else if (type.equalsIgnoreCase("awc")) {
                    typeid = 1;
                } else if (type.equalsIgnoreCase("dbo")) {
                    typeid = 2;
		} else {
                    log.log("Unknown type id " + type, Log.LOG_WARNING);
                }

                StringTokenizer rtok = new StringTokenizer(range, "-");
                int start = Integer.parseInt(rtok.nextToken());
                int finish = Integer.parseInt(rtok.nextToken());

                log.log("Creating " + type + " (type " + typeid + ") " + range, Log.LOG_INFO);
                for (int i = start; i <= finish; i++) {
                    VariableModel model = new VariableModel(this, "ND" + i);
                    switch (typeid) {
                    case 0:
                        addAgent(new RandomAgent(this, model)); break;
                        //case 1:
                        //addAgent(new AWCAgent(this, model)); break;
                        //case 2:
                        //addAgent(new DBOAgent(this, model)); break;
                    }
                }
            }
        } else {
            log.log("Could not find agent description property " + prop, Log.LOG_WARNING);
        }
    }

    /**
     */
    protected void pulseBegin() {

        super.pulseBegin();

        // Look for a new description
        String prop = getMetAgentName() + AGENTS_DESCRIPTION;
        String agentsdesc = (String)getProperty(prop);
        if (agentsdesc != null) {
            if (lastdescription == null || !lastdescription.equals(agentsdesc)) {
                createAllAgents();
            }
        }
    }

    /**
     * The main function
     */
    public static void main(String[] args) {

        try {
            FarmMetAgent agent = new SATMetAgent(args);
            agent.run();
        } catch (RemoteException e) {
            System.err.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }
}
