package farm.sat.plugins.display;
/**
 * SATPanel.java
 *
 *
 * Created: Wed Jul 10 14:47:27 2002
 *
 * @author Roger Mailler
 * @version
 */

import java.awt.*;
import javax.swing.*;
import java.awt.event.*;
import java.util.*;
import java.awt.geom.*;
import java.awt.print.*;

public class SATPanel extends JPanel
    implements MouseListener, MouseMotionListener, Printable{

    SATDisplayFrame GDF = null;
    int panWidth, panHeight;
    int envWidth, envHeight;
    Rectangle viewPort;
    int zoomLevel = 1;
    int baseNodeSize = 6;
    NodeVisual dragging;
    ArrayList nodes;
    int dragx, dragy;

    // Preferences
    boolean shownegot = true;
    boolean nodetext = false;
    boolean showconsttext = false;
    boolean showconstlinks = true;
    boolean showlinking = false;


    public SATPanel(SATDisplayFrame gdf) {
	GDF = gdf;
	panWidth = 600;
        panHeight = 600;
	viewPort = new Rectangle(0,0,panWidth, panHeight);
	addMouseListener(this);
	addMouseMotionListener(this);
	//nodes = GDF.getNodes();

    }

    public void setShowNegotiations(boolean b) { shownegot = b; }
    public boolean getShowNegotiations() { return shownegot; }
    public void setShowNodeText(boolean b) { 
	nodetext = b;
	for(Iterator i = nodes.iterator();i.hasNext();){
	    NodeVisual TMV = (NodeVisual) i.next();
	    TMV.setTextMode(b);
	}
    }
    public boolean getShowNodeText() { return nodetext; }
    public void setShowConstText(boolean b) { showconsttext = b; }
    public boolean getShowConstText() { return showconsttext; }
    public void setShowConstLinks(boolean b) { showconstlinks = b; }
    public boolean getShowConstLinks() { return showconstlinks; }
    public void setShowLinking(boolean b) { showlinking = b; }
    public boolean getShowLinking() { return showlinking; }


    public void mouseEntered(MouseEvent e) {}
    public void mouseExited(MouseEvent e) {}
    public void mousePressed(MouseEvent e) {	
	//find out if it is in a box and which one it is
	//String oldName = (String) order.elementAt( order.size()-1);
	//NodeVisual oldSel = (NodeVisual) nodes.get(oldName);
	for(int i = nodes.size()-1;i>-1;i--){
	    NodeVisual TMV = (NodeVisual) nodes.get(i);
	    if(TMV.isInside(e.getX(), e.getY(), zoomLevel)){
		dragx = e.getX()/zoomLevel - TMV.getX();
		dragy = e.getY()/zoomLevel - TMV.getY();
		dragging = TMV;
		nodes.remove(i);
		nodes.add(TMV);
		TMV.setSelected();
		break;
	    }
	}
    }
    public void mouseMoved(MouseEvent e) {}

    public void mouseReleased(MouseEvent e) {
	if(dragging !=null){

	    dragging.setX(boundPosition(e.getX()/zoomLevel- dragx, 0, envWidth));
	    
	    dragging.setY(boundPosition(e.getY()/zoomLevel- dragy, 0, envHeight));
	    dragging = null;
	}
    }
    public void mouseDragged(MouseEvent e) {
	if(dragging !=null){
	    dragging.setX(boundPosition(e.getX()/zoomLevel- dragx, 0, envWidth));

	    dragging.setY(boundPosition(e.getY()/zoomLevel- dragy, 0, envHeight));
	    repaint();
	}
    }


    public void mouseClicked(MouseEvent e) {
	
	if((e.getModifiers() & e.BUTTON1_MASK) !=0){
	//find out if it is in a box and which one it is
	    for(int i = nodes.size()-1;i>-1;i--){
		NodeVisual TMV = (NodeVisual) nodes.get(i);
		if(TMV.isInside(e.getX(), e.getY(), zoomLevel)){
		    nodes.remove(i);
		    nodes.add(TMV);
		    repaint();
		    return;
		}
	    }

	    Dimension d = getSize();
	    float relOrgX = (float) e.getPoint().x/(float)d.width;
	    float relOrgY = (float) e.getPoint().y/(float)d.height;
	    //System.err.println("click " + e.getPoint().x + " relOrgX " + relOrgX);

	    if(zoomLevel < 32) {
		zoomLevel *=2;
		float newOrgX = (float) (viewPort.getX() + viewPort.getWidth() * relOrgX);
		float newOrgY = (float) (viewPort.getY() + viewPort.getHeight() * relOrgY);
		float newEnvWidth = (float) envWidth/(float) zoomLevel;
		float newEnvHeight = (float) envHeight/(float) zoomLevel;

		newOrgX -=  newEnvWidth/2f;
		newOrgY -= newEnvHeight/2f;
		
		//check the bounds
		if(newOrgX <0) newOrgX =0f;
		if(newOrgX +  newEnvWidth > envWidth) 
		    newOrgX = envWidth- newEnvWidth;
		if(newOrgY <0) newOrgY =0f;
		if(newOrgY +  newEnvHeight > envHeight) 
		    newOrgY = envHeight- newEnvHeight;

		invalidate();
		GDF.getContentPane().validate();
	    }

	}
	else if ((e.getModifiers() & e.BUTTON3_MASK) !=0){
	    if(zoomLevel > 1){
		zoomLevel /=2;
		float newOrgX =  (float) (viewPort.getX() - viewPort.getWidth()/2.0);
		float newOrgY = (float) (viewPort.getY() - (float) viewPort.getHeight()/2.0f);

		float newEnvWidth = (float) envWidth/(float) zoomLevel;
		float newEnvHeight = (float) envHeight/(float) zoomLevel;

		//check the bounds
		if(newOrgX <0) newOrgX =0f;
		if(newOrgX +  newEnvWidth > envWidth) 
		    newOrgX = envWidth- newEnvWidth;
		if(newOrgY <0) newOrgY =0f;
		if(newOrgY +  newEnvHeight > envHeight) 
		    newOrgY = envHeight- newEnvHeight;

		invalidate();
		GDF.getContentPane().validate();
	    }
	}
	//System.err.println("zoom " + zoomLevel);
	//System.err.println("port " + viewPort.toString());
    }
    public void setEnvWidth(int w){
	envWidth = w;
	viewPort.setSize(w, (int) viewPort.getHeight());
    }
    public void setEnvHeight(int h){
	envHeight = h;
	viewPort.setSize((int)viewPort.getWidth(), h);
    }

    public void paintComponent(Graphics g) {
        Graphics2D g2 = (Graphics2D)g;
	super.paintComponent(g);
	Color c = g.getColor();

	if (envWidth == 0) return;


	g.clearRect(0, 0, getWidth(), getHeight());
	if(nodes == null){
	    nodes = GDF.getNodes();
	    if(nodes == null)
		return;
	}
	
	if(showconstlinks){
	    resetDones(nodes);
	    for(int i = 0;i<nodes.size();i++){
		NodeVisual TMV = (NodeVisual) nodes.get(i);
		ArrayList links = TMV.getLinks();
		
		
		for(int j = 0;j<links.size();j++){
		    NodeLink ln = (NodeLink) links.get(j);
		    //has it been painted?
		    
		    NodeVisual oTMV = null;
		    for(Iterator k = nodes.iterator();k.hasNext();){
			NodeVisual n = (NodeVisual) k.next();
			if(n.getName().equals(ln.getWho())){
			    oTMV = n;
			    break;
			}
		    }
		    
		    if(!oTMV.done()){
			//other node not drawn, so draw the link
			ln.resetHot();
			if(oTMV.getNodeColor() == TMV.getNodeColor()){
			    ln.setHot();
			}
			ln.draw(g, TMV.getX()*zoomLevel, TMV.getY()*zoomLevel, oTMV.getX()*zoomLevel, oTMV.getY()*zoomLevel, showconsttext);
		    }
		}
		TMV.setDone();
	    }
	}

	if(showlinking){
	    //reset the color
	    g.setColor(Color.blue);
	    resetDones(nodes);
	    for(int i = 0;i<nodes.size();i++){
		NodeVisual TMV = (NodeVisual) nodes.get(i);
		ArrayList links = (ArrayList) GDF.getProperty(TMV.getName() + ":links");
		if(links == null) continue;
		
		for(int j = 0;j<links.size();j++){
		    String who = (String) links.get(j);
		    NodeVisual oTMV = null;
		    for(Iterator k = nodes.iterator();k.hasNext();){
		    	NodeVisual n = (NodeVisual) k.next();
		    	if(n.getName().equals(who)){
		    	    oTMV = n;
		    	    break;
		    	}
		    }
		    
		    if(!oTMV.done()){
			//other node not drawn, so draw the link	
			g.drawLine(TMV.getX()*zoomLevel, TMV.getY()*zoomLevel, oTMV.getX()*zoomLevel, oTMV.getY()*zoomLevel);
		    }
		}
		TMV.setDone();
	    }
	    g.setColor(c);
	}




	for(int i = 0;i<nodes.size();i++){
	    NodeVisual TMV = (NodeVisual) nodes.get(i);

	    boolean val = ((Boolean) GDF.getProperty(TMV.getName()+":value")).booleanValue();

	    TMV.setNodeColor((val?1:0));
	    TMV.draw(g, zoomLevel);
	}
    } 


    private void resetDones(ArrayList nodes){
	for(Iterator i = nodes.iterator();i.hasNext();){
	    NodeVisual TMV = (NodeVisual) i.next();
	    TMV.resetDone();
	    TMV.resetLinks();
	}
    }

    public int print(Graphics g, PageFormat pageFormat, int pageIndex) {
        if(pageIndex ==0){
            Graphics2D g2 = (Graphics2D) g;
	    
            g2.setBackground(Color.white);
            double pageHeight = pageFormat.getImageableHeight();
            double pageWidth = pageFormat.getImageableWidth();
	    
            //figure out the scaling for the image
            Dimension d = getPreferredSize();
            double xScale = pageWidth/d.getWidth();
            double yScale = pageHeight/d.getHeight();
	    
            AffineTransform SAF = g2.getTransform();
            AffineTransform AF = new AffineTransform(SAF);
	    
            AF.translate(pageFormat.getImageableX(),
                         pageFormat.getImageableY());
	    
            if(xScale < yScale)
		AF.scale(xScale, xScale);
            else
                AF.scale(yScale, yScale);
	    
            g2.setTransform(AF);
            paintComponent(g2);
            g2.setTransform(SAF);
            return PAGE_EXISTS;
        }
        else{
            return NO_SUCH_PAGE;
        }
    }



    public void updateDisplay(){
	//System.err.println("Redraw");
	repaint();
    }

    public Dimension getPreferredSize() {
	//find the maximum extentx and extent y and return them 
	//or minimum of 600  for each
	int maxExX = 600;
	int maxExY = 600;
	return new Dimension(maxExX*zoomLevel, maxExY*zoomLevel);
    }
    
    private int boundPosition(int pos, int low, int high){
	if(pos > low){
	    if(pos <= high){ 
		return pos;
	    } else {
		return high;
	    }
	}
	else{
	    return low;
	}

    }
    
    public Dimension getMinimumSize() {return getPreferredSize(); }
    public Dimension getMaximumSize() {return getPreferredSize(); }

    private float Closest(double num, int precision){
	double scale;
	
	scale = Math.pow(10, precision);
	return (float) (Math.round(num / scale) * scale);
    }

} // SATPanel
